! Module      : ommag_mod
!
! Authors     : Tim Harries (MSSL)
!             : Phil Williams (MSSL)
!
! Description : Convert from count-rates to magnitudes in the instrumental system.
!
! Comments    :
!
! $Id: module.f90,v 1.4 1997/12/18 16:01:10 gvacanti Exp $
!
! History : Created  : March 98 - TJH
!           Modified : 20 May 99 - Spring clean of code.
!           Modified : 22/8/2000 (CB) - added call to CAL_omGetColorTrans
!                      to convert instrumental magnitudes to standard magnitudes.
!           modified : 19/9/2000 (AMK) - corrected call to CAL_omPhotoMag. Parameters 
!                      were set incorrectly.
!    4th. October 2000 - AMK
!    Replace the call getFIlterID in mssllib by the OAL call filterIdToInt
!    16th May 2001 (CB) : To comply with the CAL definitions of magnitudes, etc,
!                         the counts (after having been corrected for coincidence
!                         losses, etc, have to be scaled from the nominal 12 pixel
!                         aperture radius to the CAL specified radius for the
!                         particular filter (12 for U,B and V, 35 for UV filters).
!                         Hence no correction for optical filters.
!   7th June 2001 (CB) :  Modified to read in the source-list column of source
!                         aperture-radii produced by omdetect. If a particular
!                         radius is less than the 12-pixel value required for the
!                         correction of coincidence losses, etc, then the cal routine
!                         "CAL_omPsfEncircledEnergy" is used to extrapolate the counts.
!                         A further extrapolation is done for UV filters- from 12 pixels
!                         to 35 pixels.  
!
!
!  6th of july 2001 VNY   small correction has been made (PI) 
!  5tn Decemmer 2001 (CB) BINBPE now read in for calculating the frame-time
!  18thJanuary 2002 (CB) 0.02 magnitude offset added to take account of systematic errors
!  14 th May 2002 (CB) Extra columns added to store the scorrected count rates and some
!                      key words added to source-list file 
!  16th August 2002 (CB) Since omdetect has been modified to do the coincidence-loss
!   correction ommag has been modified not to do this one.
!********************************************************************************
module ommag_mod

   use param
   use cal
   use dal
   use errorHandling
   use frameTime_mod
   use omallocation_mod
   use omutility_mod
   use ommagnitude_mod
   implicit none

   private

   public ommag

contains

!---------------------------------------------------------------------
subroutine ommag
!---------------------------------------------------------------------
    character(255) :: set                                         ! OM OSW Source List filename.
    character(8) :: dataMode                                      ! Input file data mode.
    character(9) :: filter                                        ! Keywords.

    type(DataSetT) :: inDataSetH                                      ! Input dataset handle.
    type(TableT) :: inTableH                                          ! Input table handle.
    integer, parameter :: noColumns = 7
    type(ColumnT), dimension(noColumns) :: inColumnH                          ! Input image column handle.

    integer(int32) :: nSources                                   ! Number of sources.
    integer(int32) :: ccdMode                                    ! OM CCD data mode.
   
    integer(int32) :: i                                          ! Count.
    integer(int32) :: filterIdInt
    integer(int8), dimension(:), pointer :: extended

    real(double) :: corrRate                                      ! Corrected count rate.
    real(double) :: corrRateErrPlus                               ! Corrected rate error + (asymmetric).
    real(double) :: corrRateErrMinus                              ! Corrected rate error - (asymmetric).

    real(double) :: corrMag                                      ! Corrected magnitude.
    real(double) :: corrMagErrPlus                               ! Corrected magnitude error + (asymmetric).
    real(double) :: corrMagErrMinus                              ! Corrected magnitude error - (asymmetric).
    real(single) :: countFrameRatio = 0.5
    real(single) :: frameRate 
    real(single), dimension(:), pointer :: rawRate               ! Raw rates pointer.
    real(single), dimension(:), pointer :: rawRateErr            ! Raw rates error pointer.
    real(single), dimension(:), pointer :: correctedRate         ! Corrected rates pointer.
    real(single), dimension(:), pointer :: correctedRateErr      ! Corrected rates error pointer.


    real(double), parameter :: PI = 3.141592653589793_double  

    real(single), dimension(:), pointer :: sourceXCoord          ! Source x-coordinate.
    real(single), dimension(:), pointer :: sourceYCoord          ! Source y-coordinate. 
    
    real(single), dimension(:), pointer :: significance          ! source significance


    real(single), dimension(:), pointer :: rawBackGroundRate        ! Source background raw rate pointer.
    real(single), dimension(:), pointer :: correctedBackGroundRate  ! Source background corrected rate pointer.

    real(single), dimension(:), pointer :: psf2Correction          ! psf correction

    real(single), dimension(:), pointer :: mag                   ! Magnitude pointer.
    real(single), dimension(:), pointer :: magErr                ! Magnitude error pointer.

    real(double) :: frameTime                                    ! OM CCD frame time.
    real(double) :: deadFrac = 0.0                               ! Dead time in units of CCD frame time.
   
    real(double)              :: detectionLimitCountRate         ! The detection limit count rate (counts/second)
    real(double)              :: detectionLimitMagnitude         ! The detection limit magnitude
    real(single) :: error = 0.0  
    real(single) :: calApertureRadius = 12.0          ! Radius used for coincidence losses, etc   
    real(single) :: photometryApertureRadius          ! Radius used for converting counts to magnitudes
    logical :: scaleCounts = .false.
    real(double) :: countsRatio = 1.0_double
    real(double) :: scaleCountsRatio = 1.0_double
    real(double) :: exposure   
    integer           :: position
    real(single) :: nullValue = -999.999  
    real(double) :: deadFractionCorrection
    real(single) :: correctedBackgroundRateErr
    !************************
    ! Read in parameters.
    !************************
    set = stringParameter("set")

    

    ! Read in the OM OSW Source List file.
    inDataSetH = dataSet(set, modify)
   
         
    if (.not. hasBlock(inDataSetH, "SRCLIST")) then
        !************************************
        ! Incompatible file - can't continue
        !************************************
        write (errstr, *) "Could not find any compatibile blocks in set ", set
        call fatal("badFileType")
    end if

    !**************************************************************
    ! Read in the detection limit count rate from the PRIMARY header
    !**************************************************************
    detectionLimitCountRate = real(real32Attribute(inDataSetH, "CRLIMIT"))
           
    ! Get a handle on the source list binary table extension.
    inTableH = table(inDataSetH, "SRCLIST")



    !******************************************************************
    ! Get the number of rows in the source list binary table extension.
    !******************************************************************
    nSources = numberOfRows(inTableH)

    

    if (nSources .le. 0) then
        call release(inTableH)
        call release(inDataSetH)
        call warning("No sources in source-list file")
        stop
    end if        
 

    ! Read in the source list table extension keywords.
    dataMode = stringAttribute(inDataSetH, "DATAMODE")
    filter = stringAttribute(inDataSetH, "FILTER")
    exposure = real32Attribute(inDataSetH, "EXPOSURE")

  
    frameTime = real64Attribute(inDataSetH, "FRAMTIME") / 1000.0
    deadFrac = real64Attribute(inDataSetH, "DEADFRAC")



    deadFractionCorrection = 1.0_double / (1.0_double - deadFrac)

    frameRate = 1.0 / frameTime

    !********************
    ! Get the ccd mode.
    !********************
    if (trim(dataMode) .eq. "IMAGING") then
        ccdMode = IMAGE
    else
        ccdMode = FAST
    end if

    !**************************************
    ! Get the filter Id integer value
    !**************************************
    filterIdInt = filterIdToInt(filter)

    !*******************
    ! Set the CAL state
    !********************
    call CAL_setState(OM, ccdModeId=ccdMode, filterId = filterIdInt)


    !*************************************************************
    ! 17th May 2001 - Due to the changes to the photometry in
    ! omdetect, now need to scale the source counts from the 12
    ! pixel aperture radius to that specified by the CAL.
    ! For U, B and V there will be no correction.
    !************************************************************

    photometryApertureRadius = CAL_omGetApertureRadius()
 

    if (photometryApertureRadius .gt. calApertureRadius) then
        !**********************************************
        ! Remember that counts have to be extrapolated.
        !**********************************************
        scaleCounts = .true.
        scaleCountsRatio = (photometryApertureRadius / calApertureRadius) ** 2 
    end if



    !******************************************************
    ! Calculate the instrumental limiting magnitude from
    ! the detection limit count rate
    !******************************************************
    if (detectionLimitCountRate .gt. 0.0) then      


        corrRate = CorrectForCoincidenceLoss(detectionLimitCountRate, error, frameTime, &
                  corrRateErrMinus,  corrRateErrPlus)
               
        !************************************************
        ! Correct for dead-time
        !************************************************
        corrRate = corrRate * deadFractionCorrection 
        corrRateErrMinus = corrRateErrMinus * deadFractionCorrection 
        corrRateErrPlus = corrRateErrPlus * deadFractionCorrection 

        if (corrRate .gt. 0.0) then
            !**********************************************************
            ! May need to scale the counts from 12 to 35 pixels for uv
            !*********************************************************
            if (scaleCounts) then
                corrRate = corrRate * scaleCountsRatio
                corrRateErrMinus = corrRateErrMinus * scaleCountsRatio
                corrRateErrPlus = corrRateErrPlus * scaleCountsRatio

            end if

     
            call CAL_omPhotoMagnitude(corrRate, detectionLimitMagnitude, corrRateErrMinus, &
                                   corrRateErrPlus, corrMagErrMinus, corrMagErrPlus)
        else
            detectionLimitMagnitude = 0.0d0
        end if
    else
        detectionLimitmagnitude = 0.0
    end if

    call setAttribute(inDataSetH, "MAGLIMIT", real(detectionLimitMagnitude), &
             "mag", "detection limiting magnitude (instrumental)")


    !**************************************************
    ! Get the data from the source list file
    ! The rates have been corrected for significance-losses
    ! by omdetect- referred to as RAW_RATES
    !**************************************************

    rawRate => GetRealColumn(inTableH, "CORR_RATE", inColumnH(1))
    rawRateErr => GetRealColumn(inTableH, "CORR_RATE_ERR", inColumnH(2))


    mag => AllocateSingle(0, nSources - 1, "mag")
    magErr => AllocateSingle(0,  nSources - 1, "magErr")

 
   
    rawBackGroundRate => GetRealColumn(inTableH, "CORR_BACKGROUND_RATE", inColumnH(3))


    sourceXCoord => GetRealColumn(inTableH, "XPOS", inColumnH(4))


    sourceYCoord => GetRealColumn(inTableH, "YPOS", inColumnH(5))


  
    significance => GetRealColumn(inTableH, "SIGNIFICANCE", inColumnH(6))


    inColumnH(7) = column(inTableH, "EFLAG", modify)


    extended => int8Data(inColumnH(7))
    

    !******************************************************************
    ! Loop through all the sources- 
    ! Correct each count-rate for the effect of detector 
    ! lifetime and ageing, then calculate the magnitude.
    !******************************************************************
   
      
    !****************************************************
    ! Arrays for storing the corrected count rates, etc
    !**************************************************** 
    correctedRate => AllocateSingle(0, nSources - 1, "correctedRate")  
    correctedRate = nullValue
    correctedRateErr => AllocateSingle(0, nSources - 1, "correctedRateErr")  
    correctedRateErr = nullValue
 
    correctedBackgroundRate => AllocateSingle(0, nSources - 1, "correctedBackgroundRate")  
    correctedBackgroundRate = nullValue
 
  
     
    psf2Correction => AllocateSingle(0, nSources - 1, "psf2Correction")  
    psf2Correction = nullValue

    !********************************************************
    ! Loop through the sources, do the dead-time correction
    ! and then compute the instrumental magnitude
    !********************************************************
!write(*, *) "i  XCoord  YCoord  rawRate  correctedRate rawBackgroundrate  correctedBackgroundRate magnitude error" 

    correctedRate = rawRate
    correctedRateErr = rawRateErr
    correctedBackgroundRate = rawBackgroundRate
    mag = nullValue
    magErr = nullValue

 
   do i = 0, nSources - 1

!write(*, 100) i,  sourceXCoord(i),  sourceYCoord(i),  rawRate(i),  correctedRate(i), &
!                 rawBackgroundrate(i),  correctedBackgroundRate(i), mag(i), magerr(i) 

       
               
          
        call ComputeMagnitude(correctedRate(i), correctedRateErr(i), &
                            correctedBackgroundRate(i),  correctedBackgroundRateErr, &
                            sourceXCoord(i), sourceYCoord(i),             &
                            scaleCounts, extended(i), &
                            photometryApertureRadius, calApertureRadius, &
                            frameTime, deadFrac, &
                            mag(i), magErr(i), psf2Correction(i))


        significance(i) = ComputeSourceSignificance(correctedRate(i) * real(exposure), &
                real(exposure) * correctedBackgroundRate(i))

     
!write(*, 100) i,  sourceXCoord(i),  sourceYCoord(i),  rawRate(i),  correctedRate(i), &
!                 rawBackgroundrate(i),  correctedBackgroundRate(i), mag(i), magerr(i) 

!100 format(i6, 2f8.2, 4f12.2, 2f8.2)
    end do

   


    !*************************
    ! Release handles on data
    !***********************

    do i = 1, noColumns
        call release(inColumnH(i))
    end do

  
    !**************************************************
    ! Add the columns for the corrected count rates,
    ! after the significance column
    !**************************************************
    position = columnNumber(inTableH, "SIGNIFICANCE")
    call AddReal32Column(inTableH, correctedRate, nSources, "CORR_RATE", &
            "counts/sec",  "Corrected for coincidence losses, deadfraction, etc", nullValue, position)

    call AddReal32Column(inTableH, correctedRateErr, nSources, "CORR_RATE_ERR", &
            "counts/sec", "Corrected for coincidence losses, deadfraction, etc" , nullValue, position + 1)

  
 
    call AddReal32Column(inTableH, psf2Correction, nSources, "PSF2_CORR", &
            "%", "PSF2 correction (12pixels to cal radius)", nullValue, position + 5)
    

    call AddReal32Column(inTableH, correctedBackgroundRate, nSources, "CORR_BACKGROUND_RATE", &
            "counts/sec", "Corrected for coincidence losses, deadfraction, etc", nullValue, position + 5)

    !****************************************************************
    ! Add the magnitudes
    !**************************************************************** 
    position = columnNumber(inTableH, "COINCLOSS_CORR") + 1
    
    call AddReal32Column(inTableH, mag, nSources, "MAG", &
            "mag", "Instrumental magnitude", nullValue, position)
    deallocate(mag)

    !****************************************************************
    ! Add the magnitude errors
    !**************************************************************** 
    call AddReal32Column(inTableH, magErr, nSources, "MAGERR", &
            "mag", "Instrumental magnitude error", nullValue, position + 1)
    
    deallocate(magErr)

    call release(inTableH)


    call release(inDataSetH)

    deallocate(correctedRate)
    deallocate(correctedRateErr)
    deallocate(correctedBackgroundRate)

    deallocate(psf2Correction)
   
end subroutine ommag
!*********************************************************
! Function CorrectForCoincidenceLoss
! Corrects the raw count rate for coincidence losses  
!*********************************************************
real(double) function CorrectForCoincidenceLoss(rawCountRate, & 
                  countRateError, frameTime, &
                  countRateErrorMinus, countRateErrorPlus)
    real(double), intent(in)     :: rawCountRate
    real(single), intent(in)     :: countRateError
    real(double), intent(out)    :: countRateErrorMinus
    real(double), intent(out)    :: countRateErrorPlus
 
    real(double), intent(in) :: frameTime
    
    
    logical :: isExtended
    type(ValueAsymmetricErrorT) :: rawRate
    type(ValueAsymmetricErrorT) :: correctedRate

    !*************************************************************
    ! NB frame time is in milli-seconds
    !*************************************************************
 
    if (rawCountRate .gt. (1000.0_double / frameTime)) then
        call warning("Count-rate>frame-rate", "Coincidence-loss correction ignored")

    else
        isExtended = .false.
    end if
           
 
    rawRate = ValueAsymmetricErrorT(rawCountRate, dble(countRateError), dble(countRateError))


    correctedRate = CAL_omPhotoNatural(rawRate, frameTime, 0.0_double, 0.0_double, isExtended, .TRUE.)


    !****************************************************************
    ! If the count rate is greater than the frame rate then the above
    ! routine cannot do the correction- instead it sets the corrected 
    ! count rate to the largest number that can be stored on the computer. 
    !*****************************************************************

    if (CorrectedRate%value .lt. 1e10) then
        CorrectForCoincidenceLoss = real(CorrectedRate%value)
        
        countRateErrorMinus = correctedRate%errorMinus
     
        countRateErrorPlus  = correctedRate%errorPlus

    else
        CorrectForCoincidenceLoss = rawCountRate
    end if

end function CorrectForCoincidenceLoss



!****************************************
! Function GetRealColumn
! Returns an array containing the data
! in the named column of the given table
!***************************************
Function GetRealColumn(inTableH, name, columnH)  result(dataPtr)  
    type(TableT), intent(in) :: inTableH
    character(*), intent(in) :: name
    type(ColumnT), intent(inout) :: columnH
    real(single), dimension(:), pointer :: dataPtr                ! Magnitude error pointer.
   
    columnH = column(inTableH, name, modify)
    dataPtr => real32Data(columnH)

end function GetRealColumn
!***********************************************
! ComputeSourceSignificance
! Computes the source significance and returns
! the value via the name
!***********************************************
real(single) function ComputeSourceSignificance(sourceCounts, backGroundCounts)
    real(single), intent(in) :: sourceCounts     ! The background-subtracted source counts
    real(single), intent(in) :: backGroundCounts ! The background counts

    if (backGroundCounts > 0.0) then
        ComputeSourceSignificance = sourceCounts  / sqrt(backGroundCounts)
    else
        ComputeSourceSignificance = sourceCounts
    end if
end function ComputeSourceSignificance




end module ommag_mod












